@@ -6,8 +6,8 @@ RSpec::Matchers.define :fail_with do |message|
   end
 end
 
-def post_xml(xml=:example_response)
-  post "/auth/saml/callback", {'SAMLResponse' => load_xml(xml)}
+def post_xml(xml=:example_response, opts = {})
+  post "/auth/saml/callback", opts.merge({'SAMLResponse' => load_xml(xml)})
 end
 
 describe OmniAuth::Strategies::SAML, :type => :strategy do
@@ -17,7 +17,9 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
   let(:saml_options) do
     {
       :assertion_consumer_service_url     => "http://localhost:9080/auth/saml/callback",
+      :single_logout_service_url          => "http://localhost:9080/auth/saml/slo",
       :idp_sso_target_url                 => "https://idp.sso.example.com/signon/29490",
+      :idp_slo_target_url                 => "https://idp.sso.example.com/signoff/29490",
       :idp_cert_fingerprint               => "C1:59:74:2B:E8:0C:6C:A9:41:0F:6E:83:F6:D1:52:25:45:58:89:FB",
       :idp_sso_target_url_runtime_params  => {:original_param_key => :mapped_param_key},
       :name_identifier_format             => "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress",
@@ -39,11 +41,11 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
       end
 
       it 'should get authentication page' do
-        last_response.should be_redirect
-        last_response.location.should match /https:\/\/idp.sso.example.com\/signon\/29490/
-        last_response.location.should match /\?SAMLRequest=/
-        last_response.location.should_not match /mapped_param_key/
-        last_response.location.should_not match /original_param_key/
+        expect(last_response).to be_redirect
+        expect(last_response.location).to match /https:\/\/idp.sso.example.com\/signon\/29490/
+        expect(last_response.location).to match /\?SAMLRequest=/
+        expect(last_response.location).not_to match /mapped_param_key/
+        expect(last_response.location).not_to match /original_param_key/
       end
     end
 
@@ -53,11 +55,11 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
       end
 
       it 'should get authentication page' do
-        last_response.should be_redirect
-        last_response.location.should match /https:\/\/idp.sso.example.com\/signon\/29490/
-        last_response.location.should match /\?SAMLRequest=/
-        last_response.location.should match /\&mapped_param_key=original_param_value/
-        last_response.location.should_not match /original_param_key/
+        expect(last_response).to be_redirect
+        expect(last_response.location).to match /https:\/\/idp.sso.example.com\/signon\/29490/
+        expect(last_response.location).to match /\?SAMLRequest=/
+        expect(last_response.location).to match /\&mapped_param_key=original_param_value/
+        expect(last_response.location).not_to match /original_param_key/
       end
     end
 
@@ -71,17 +73,17 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
         %w(foo.example.com bar.example.com).each do |host|
           get "https://#{host}/auth/saml"
 
-          last_response.should be_redirect
+          expect(last_response).to be_redirect
 
           location = URI.parse(last_response.location)
           query = Rack::Utils.parse_query location.query
-          query.should have_key('SAMLRequest')
+          expect(query).to have_key('SAMLRequest')
 
           request = REXML::Document.new(Base64.decode64(query['SAMLRequest']))
-          request.root.should_not be_nil
+          expect(request.root).not_to be_nil
 
           acs = request.root.attributes.get_attribute('AssertionConsumerServiceURL')
-          acs.to_s.should == "https://#{host}/auth/saml/callback"
+          expect(acs.to_s).to eq "https://#{host}/auth/saml/callback"
         end
       end
     end
@@ -93,7 +95,7 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
     let(:xml) { :example_response }
 
     before :each do
-      Time.stub(:now).and_return(Time.utc(2012, 11, 8, 20, 40, 00))
+      allow(Time).to receive(:now).and_return(Time.utc(2012, 11, 8, 20, 40, 00))
     end
 
     context "when the response is valid" do
@@ -102,17 +104,21 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
       end
 
       it "should set the uid to the nameID in the SAML response" do
-        auth_hash['uid'].should == '_1f6fcf6be5e13b08b1e3610e7ff59f205fbd814f23'
+        expect(auth_hash['uid']).to eq '_1f6fcf6be5e13b08b1e3610e7ff59f205fbd814f23'
       end
 
       it "should set the raw info to all attributes" do
-        auth_hash['extra']['raw_info'].all.to_hash.should == {
+        expect(auth_hash['extra']['raw_info'].all.to_hash).to eq(
           'first_name'   => ['Rajiv'],
           'last_name'    => ['Manglani'],
           'email'        => ['user@example.com'],
           'company_name' => ['Example Company'],
           'fingerprint'  => saml_options[:idp_cert_fingerprint]
-        }
+        )
+      end
+
+      it "should set the response_object to the response object from ruby_saml response" do
+        expect(auth_hash['extra']['response_object']).to be_kind_of(OneLogin::RubySaml::Response)
       end
     end
 
@@ -124,17 +130,17 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
       end
 
       it "should set the uid to the nameID in the SAML response" do
-        auth_hash['uid'].should == '_1f6fcf6be5e13b08b1e3610e7ff59f205fbd814f23'
+        expect(auth_hash['uid']).to eq '_1f6fcf6be5e13b08b1e3610e7ff59f205fbd814f23'
       end
 
       it "should set the raw info to all attributes" do
-        auth_hash['extra']['raw_info'].all.to_hash.should == {
+        expect(auth_hash['extra']['raw_info'].all.to_hash).to eq(
           'first_name'   => ['Rajiv'],
           'last_name'    => ['Manglani'],
           'email'        => ['user@example.com'],
           'company_name' => ['Example Company'],
           'fingerprint'  => 'C1:59:74:2B:E8:0C:6C:A9:41:0F:6E:83:F6:D1:52:25:45:58:89:FB'
-        }
+        )
       end
     end
 
@@ -148,6 +154,7 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
 
     context "when there is no name id in the XML" do
       before :each do
+        allow(Time).to receive(:now).and_return(Time.utc(2012, 11, 8, 23, 55, 00))
         post_xml :no_name_id
       end
 
@@ -191,38 +198,103 @@ describe OmniAuth::Strategies::SAML, :type => :strategy do
       end
 
       it "should obey attribute statements mapping" do
-        auth_hash[:info].should == {
+        expect(auth_hash[:info]).to eq(
           'first_name'   => 'Rajiv',
           'last_name'    => 'Manglani',
           'email'        => 'user@example.com',
           'name'         => nil
-        }
+        )
+      end
+    end
+
+    context "when response is a logout response" do
+      before :each do
+        saml_options[:issuer] = "https://idp.sso.example.com/metadata/29490"
+
+        post "/auth/saml/slo", {
+          SAMLResponse: load_xml(:example_logout_response),
+          RelayState: "https://example.com/",
+        }, "rack.session" => {"saml_transaction_id" => "_3fef1069-d0c6-418a-b68d-6f008a4787e9"}
+      end
+      it "should redirect to relaystate" do
+        expect(last_response).to be_redirect
+        expect(last_response.location).to match /https:\/\/example.com\//
+      end
+    end
+
+    context "when request is a logout request" do
+      before :each do
+        saml_options[:issuer] = "https://idp.sso.example.com/metadata/29490"
+        post "/auth/saml/slo", {
+          "SAMLRequest" => load_xml(:example_logout_request),
+          "RelayState" => "https://example.com/",
+        }, "rack.session" => {"saml_uid" => "username@example.com"}
+      end
+
+      it "should redirect to logout response" do
+        expect(last_response).to be_redirect
+        expect(last_response.location).to match /https:\/\/idp.sso.example.com\/signoff\/29490/
+        expect(last_response.location).to match /RelayState=https%3A%2F%2Fexample.com%2F/
+      end
+    end
+
+    context "when sp initiated SLO" do
+      def test_default_relay_state(static_default_relay_state = nil, &block_default_relay_state)
+        saml_options["slo_default_relay_state"] = static_default_relay_state || block_default_relay_state
+        post "/auth/saml/spslo"
+
+        expect(last_response).to be_redirect
+        expect(last_response.location).to match /https:\/\/idp.sso.example.com\/signoff\/29490/
+        expect(last_response.location).to match /RelayState=https%3A%2F%2Fexample.com%2F/
+      end
+
+      it "should redirect to logout request" do
+        test_default_relay_state("https://example.com/")
+      end
+
+      it "should redirect to logout request with a block" do
+        test_default_relay_state do
+          "https://example.com/"
+        end
+      end
+
+      it "should redirect to logout request with a block with a request parameter" do
+        test_default_relay_state do |request|
+          "https://example.com/"
+        end
+      end
+
+      it "should give not implemented without an idp_slo_target_url" do
+        saml_options.delete(:idp_slo_target_url)
+        post "/auth/saml/spslo"
+
+        expect(last_response.status).to eq 501
+        expect(last_response.body).to match /Not Implemented/
       end
     end
   end
 
   describe 'GET /auth/saml/metadata' do
     before do
+      saml_options[:issuer] = 'http://example.com/SAML'
       get '/auth/saml/metadata'
     end
 
     it 'should get SP metadata page' do
-      last_response.status.should == 200
-      last_response.header["Content-Type"].should == "application/xml"
+      expect(last_response.status).to eq 200
+      expect(last_response.header["Content-Type"]).to eq "application/xml"
     end
 
     it 'should configure attributes consuming service' do
-      last_response.body.should match /AttributeConsumingService/
-      last_response.body.should match /first_name/
-      last_response.body.should match /last_name/
-      last_response.body.should match /Required attributes/
+      expect(last_response.body).to match /AttributeConsumingService/
+      expect(last_response.body).to match /first_name/
+      expect(last_response.body).to match /last_name/
+      expect(last_response.body).to match /Required attributes/
+      expect(last_response.body).to match /entityID/
+      expect(last_response.body).to match /http:\/\/example.com\/SAML/
     end
   end
 
-  it 'implements #on_metadata_path?' do
-    expect(described_class.new(nil)).to respond_to(:on_metadata_path?)
-  end
-
   describe 'subclass behavior' do
     it 'registers subclasses in OmniAuth.strategies' do
       subclass = Class.new(described_class)
